/******************************************************************************

  Copyright(c) 2006 Intel Corporation. All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of version 2 of the GNU General Public License as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110,
  USA

  The full GNU General Public License is included in this distribution
  in the file called LICENSE.

  Contact Information:
  James P. Ketrenos <ipw2100-admin@linux.intel.com>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR
  97124-6497

******************************************************************************/

/******************************************************************************

  To build:

	% gcc -I/usr/src/linux/include -Wall -o sim_rx wifi_tx.c

  To run:

	% ./wifi_tx --help

******************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/wait.h>

#include <asm/types.h>

#include <linux/rtnetlink.h>
#include <linux/netlink.h>
#include <linux/if.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <linux/if_arp.h>
#include <linux/llc.h>
#include <linux/sockios.h>
#include <linux/types.h>

#include <netinet/in.h>

#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>

typedef signed char s8;
typedef unsigned char u8;
typedef unsigned short u16;
typedef signed short s16;
typedef unsigned int u32;
typedef signed int s32;
typedef unsigned long long u64;
typedef signed long long s64;

/*#include <net/ieee80211_radiotap.h>*/

#ifndef ETH_P_80211_RAW
#define ETH_P_80211_RAW (ETH_P_ECONET + 1)
#define ETH_P_80211_RADIOTAP (ETH_P_80211_RAW + 1)
#endif


/*****************************************************************************
 *
 * IEEE 802.11 Declarations
 *
 * We define these here as including net/ieee80211.h has too many other
 * typedef dependencies.  Also, depending on the net/ieee80211.h on the system,
 * some of these types may not be defined.
 *
 */
#define IEEE80211_MAC_LEN       6
#define IEEE80211_3ADDR_LEN     24
#define IEEE80211_2ADDR_LEN 16
#define IEEE80211_3ADDR_LEN 24
#define IEEE80211_4ADDR_LEN 30
#define IEEE80211_FCS_LEN    4

/* Frame control field constants */
#define IEEE80211_FCTL_VERS		0x0003
#define IEEE80211_FCTL_FTYPE		0x000c
#define IEEE80211_FCTL_STYPE		0x00f0
#define IEEE80211_FCTL_TODS		0x0100
#define IEEE80211_FCTL_FROMDS		0x0200
#define IEEE80211_FCTL_DS_MSK           0x0300
#define IEEE80211_FCTL_MOREFRAGS	0x0400
#define IEEE80211_FCTL_RETRY		0x0800
#define IEEE80211_FCTL_PM		0x1000
#define IEEE80211_FCTL_MOREDATA		0x2000
#define IEEE80211_FCTL_PROTECTED	0x4000
#define IEEE80211_FCTL_ORDER		0x8000

#define IEEE80211_FTYPE_MGMT		0x0000
#define IEEE80211_FTYPE_CTL		0x0004
#define IEEE80211_FTYPE_DATA		0x0008

/* management */
#define IEEE80211_STYPE_ASSOC_REQ	0x0000
#define IEEE80211_STYPE_ASSOC_RESP 	0x0010
#define IEEE80211_STYPE_REASSOC_REQ	0x0020
#define IEEE80211_STYPE_REASSOC_RESP	0x0030
#define IEEE80211_STYPE_PROBE_REQ	0x0040
#define IEEE80211_STYPE_PROBE_RESP	0x0050
#define IEEE80211_STYPE_BEACON		0x0080
#define IEEE80211_STYPE_ATIM		0x0090
#define IEEE80211_STYPE_DISASSOC	0x00A0
#define IEEE80211_STYPE_AUTH		0x00B0
#define IEEE80211_STYPE_DEAUTH		0x00C0
#define IEEE80211_STYPE_ACTION		0x00D0

/* control */
#define IEEE80211_STYPE_PSPOLL		0x00A0
#define IEEE80211_STYPE_RTS		0x00B0
#define IEEE80211_STYPE_CTS		0x00C0
#define IEEE80211_STYPE_ACK		0x00D0
#define IEEE80211_STYPE_CFEND		0x00E0
#define IEEE80211_STYPE_CFENDACK	0x00F0

/* data */
#define IEEE80211_STYPE_DATA		0x0000
#define IEEE80211_STYPE_DATA_CFACK	0x0010
#define IEEE80211_STYPE_DATA_CFPOLL	0x0020
#define IEEE80211_STYPE_DATA_CFACKPOLL	0x0030
#define IEEE80211_STYPE_NULLFUNC	0x0040
#define IEEE80211_STYPE_CFACK		0x0050
#define IEEE80211_STYPE_CFPOLL		0x0060
#define IEEE80211_STYPE_CFACKPOLL	0x0070
#define IEEE80211_STYPE_QOS_DATA        0x0080

#define IEEE80211_SCTL_FRAG		0x000F
#define IEEE80211_SCTL_SEQ		0xFFF0

struct ieee80211_hdr_3addr {
	__le16 frame_ctl;
	__le16 duration_id;
	u8 addr1[IEEE80211_MAC_LEN];
	u8 addr2[IEEE80211_MAC_LEN];
	u8 addr3[IEEE80211_MAC_LEN];
	__le16 seq_ctl;
	u8 payload[0];
} __attribute__ ((packed));

/* IEEE 802.11h Declarations */
#define IEEE80211_STYPE_ACTION 0x00D0

enum {
	IEEE80211_IE_CSA = 37,
};

enum {
	IEEE80211_ACTION_SPECTRUM_MGMT = 0,
};

enum {
	IEEE80211_ACTION_CATEGORY_CHANNEL_SWITCH = 4,
};

struct ieee80211_channel_switch {
	u8 id;
	u8 len;
	u8 mode;
	u8 channel;
	u8 count;
} __attribute__ ((packed));

struct ieee80211h_csa {
	struct ieee80211_hdr_3addr header;
	u8 category;
	u8 action;
	struct ieee80211_channel_switch channel_switch;
} __attribute__ ((packed));

struct ieee80211_deauth {
	struct ieee80211_hdr_3addr header;
	__le16 reason;
} __attribute__ ((packed));

struct ieee80211_beacon {
	struct ieee80211_hdr_3addr header;
	__le32 time_high;
	__le32 time_low;
	__le16 interval;
	__le16 capabilities;
	u8 info_element[0];
} __attribute__ ((packed));

struct ieee80211_data {
	struct ieee80211_hdr_3addr header;
	u8 payload[0];
} __attribute__ ((packed));

/*****************************************************************************/

#define VERSION "0.1"
#define COPYRIGHT "Copyright (C) 2006, Intel Corporation."
struct ieee80211_radiotap_header {
	u8 it_version;
	u8 it_pad;
	__le16 it_len;
	__le32 it_present;
};

struct ipw_rt_hdr {
	struct ieee80211_radiotap_header rt_hdr;
	u64 rt_tsf;
	u8 rt_flags;
	u8 rt_rate;
	u16 rt_channel;
	u16 rt_chbitmask;
	s8 rt_dbmsignal;
	s8 rt_dbmnoise;
	u8 rt_antenna;
	u8 payload[0];
} __attribute__ ((packed));

/* 3945 Simulation Sysfs Interface declarations */

struct sim_frame {
	struct ipw_rt_hdr rt;
	union {
		struct ieee80211_deauth deauth;
		struct ieee80211h_csa csa;
		struct ieee80211_beacon beacon;
		struct ieee80211_data data;
		u8 raw[2346];
	} type;
} __attribute__ ((packed));

#define SYSFS_DRIVER "/sys/bus/pci/drivers/ipw3945"

struct priv {
	/* configuration data */
	u8 csa_channel;
	u8 channel;
	u8 src[IEEE80211_MAC_LEN];
	u8 dst[IEEE80211_MAC_LEN];
	u8 bss[IEEE80211_MAC_LEN];
	char *ifname;
	int count;

	/* run time data */
	int app_configured; /* Set if this app configured the interface */
	int sock; /* socket used for Tx/Rx of frames */
	int size;

	/* user override flags */
	int set_no_configure;
	int set_quiet;
	int set_rtap;
	int set_ifname;
	int set_channel;
	int set_bss;
	int set_dst;
	int set_src;
	int set_type;
	int set_csa_channel;
	int set_rtap_mode;
	int set_count;
	int set_size;
};

int usage(char *argv[], int ret)
{
	if (ret)
		fprintf(stderr, "Invalid parameter %d: %s.\n\n",
			ret, argv[ret]);

	fprintf(stdout, "usage: wifi_tx [--OPTION]\n"
		"options:\n"
		"  --quiet\t\tDon't display (much) to console\n"
		"  --count=NUM\t\tNumber of packets to send (default 1)\n"
		"  --i=ifname\t\tDefault wifi0 if rtap=0 or rtap0 if rtap=1\n"
		"  --rtap=1|0\t\tEnable or disable rtap control (0)\n"
		"  --type=[N]\t\tSet frame type, where type is one of:\n"
		"    0: CSA\n"
		"    1: DEAUTH\n"
		"    2: BEACON\n"
		"    3: DATA\n"
		"  --size=bytes\tBytes to send for DATA type.  Default random: %d:2346.\n"
		"  --channel=channel\tSource channel for simulated Rx.\n"
		"  --csa-channel=channel\tTarget channel used in CSA.\n"
		"  --src=SRC_MAC\t\tSource MAC in form 00:00:00:00:00:00\n"
		"  --dst=DST_MAC\t\tDestination MAC in form 00:00:00:00:00:00\n"
		"  --bss=BSS_MAC\t\tNetwowrk MAC in form 00:00:00:00:00:00\n"
		"  --version\t\tDisplay version information\n"
		"  --help\t\tView this help message\n",
		IEEE80211_3ADDR_LEN);

	return -ret;
}

enum {
	PARAM_NONE = 0,
	PARAM_VALUE = 1,
};

int tolower(int c)
{
	return (c >= 'A' && c <= 'Z') ? c - 'A' + 'a' : c;
}

int tohex(int c)
{
	return (c >= '0' && c <= '9') ? c - '0' : c - 'a' + 0xA;
}

int parse_mac(const char *param, u8 * dst)
{
	int i;
	for (i = 0; i < IEEE80211_MAC_LEN; i++) {
		if (*param == '\0')
			break;
		dst[i] = tohex(tolower(*param++)) << 4;

		if (*param == '\0')
			break;
		dst[i] |= tohex(tolower(*param++));

		if (*param != '\0' && *param == ':')
			param++;
	}

	return (i != IEEE80211_MAC_LEN);
}

inline int param_check(const char *param, const char *value, int type)
{
	int len = strlen(value);
	int comp = strncmp(param, value, len);

	if (comp)
		return 0;

	if (type == PARAM_VALUE)	/* = */
		return (param[len] != '=') ? -1 : len + 1;

	/* no = */

	return (param[len] != '\0') ? -1 : len;
}

int version()
{
	fprintf(stdout, "wifi_tx - Raw 802.11 Packet transmission\n");
	fprintf(stdout, "%s\n", COPYRIGHT);
	fprintf(stdout, "version: " VERSION "\n");
	return 0;
}

int init_radiotap(struct priv *priv, struct ipw_rt_hdr *rt)
{
	if (!priv->set_channel) {
		fprintf(stderr, "--channel must be set.\n");
		return 0;
	}
	rt->rt_channel = priv->channel;
	return (sizeof(*rt));
}

void init_ieee80211_hdr_3addr(struct priv *priv,
			      struct ieee80211_hdr_3addr *hdr, __le16 frame_ctl)
{
	u8 *src, *dst, *bss;
	
	hdr->duration_id = 0;
	hdr->seq_ctl = 0;
	hdr->frame_ctl = frame_ctl;

	switch (frame_ctl & IEEE80211_FCTL_DS_MSK) {
	case IEEE80211_FCTL_TODS: /* To DS from a STA via AP */
		dst = hdr->addr3;
		src = hdr->addr1;
		bss = hdr->addr2;
		break;
		
	case IEEE80211_FCTL_FROMDS: /* From DS to a STA via AP */
		dst = hdr->addr1;
		src = hdr->addr3;
		bss = hdr->addr2;
		break;

	case (IEEE80211_FCTL_TODS | IEEE80211_FCTL_FROMDS): /* WDS? */
		dst = hdr->addr3;
		src = hdr->addr2;
		bss = hdr->addr1;
		break;

	case 0: /* Not leaving DS or network is in Ad-Hoc */
		dst = hdr->addr1;
		src = hdr->addr2;
		bss = hdr->addr3;
		break;
	}

	memcpy(src, priv->src, IEEE80211_MAC_LEN);
	memcpy(dst, priv->dst, IEEE80211_MAC_LEN);
	memcpy(bss, priv->bss, IEEE80211_MAC_LEN);
}

int init_beacon(struct priv *priv, struct ieee80211_beacon *frame)
{
	struct timeval tv;

	if (!priv->set_bss || !priv->set_dst || !priv->set_src) {
		fprintf(stderr, "--bss, --dst, --src, and "
			"must all be set.\n");
		return 0;
	}

	gettimeofday(&tv, NULL);

	init_ieee80211_hdr_3addr(priv, &frame->header,
				 IEEE80211_FTYPE_MGMT | 
				 IEEE80211_STYPE_BEACON);

	
	frame->time_low = (tv.tv_usec << 12) | (rand() % 0xfff);
	frame->time_high = tv.tv_sec;

	if (!priv->set_quiet)
		printf("timestamp: %08x%08x\n", 
		       (u32)frame->time_high, (u32)frame->time_low);

	frame->interval = 0x0064;
	frame->capabilities = 0x0431;

	return sizeof(*frame);
}

int init_data(struct priv *priv, struct ieee80211_data *frame)
{
	int size;
	int i;

	if (priv->set_size)
		size = priv->size;
	else
		size = rand() % (2346 - IEEE80211_3ADDR_LEN);

	if (!priv->set_bss || !priv->set_dst || !priv->set_src) {
		fprintf(stderr, "--bss, --dst, --src, and "
			"must all be set.\n");
		return 0;
	}

	init_ieee80211_hdr_3addr(priv, &frame->header,
				 IEEE80211_FTYPE_DATA |
				 IEEE80211_STYPE_DATA |
				 IEEE80211_FCTL_FROMDS);

	for (i = 0; i < size; i++)
		frame->payload[i] = rand() % 0xff;


	return size + sizeof(*frame);
}

#if 0
int init_csa(struct priv *priv, struct ieee80211h_csa *frame)
{
	if (!priv->set_bss || !priv->set_dst || !priv->set_src) {
		fprintf(stderr, "--bss, --dst, --src, and "
			"must all be set.\n");
		return 0;
	}

	init_ieee80211_hdr_3addr(priv, &frame->header,
				 IEEE80211_FTYPE_MGMT | 
				 IEEE80211_STYPE_ACTION |
				 IEEE80211_FCTL_TDS);

	frame->category = IEEE80211_ACTION_SPECTRUM_MGMT;
	frame->action = IEEE80211_ACTION_CATEGORY_CHANNEL_SWITCH;

	frame->channel_switch.id = IEEE80211_IE_CSA;
	frame->channel_switch.len = 3;

	/* New Channel */
	frame->channel_switch.channel = 
		priv->set_csa_channel ? priv->csa_channel : 40;

	/* Mode:
	 * 1 - No more Tx allowed until on the new channel
	 * 0 - No Tx rerequiement on the receiving client
	 */
	frame->channel_switch.mode = 1;

	/* Channel Switch Count:
	 * >1 - Number of beacon transmission intervals until switch.
	 * 1  - Occur immediately before next TBTT.
	 * 0  - Occur immediately
	 */
	frame->channel_switch.count = 0;

	return sizeof(*frame);
}
#endif

int init_csa(struct priv *priv, struct ieee80211_beacon *frame)
{
	int size;
	struct ieee80211_channel_switch *csa;

	if (!priv->set_bss || !priv->set_dst || !priv->set_src) {
		fprintf(stderr, "--bss, --dst, --src, and "
			"must all be set.\n");
		return 0;
	}

	size = init_beacon(priv, (void *)frame);
	csa = (void *)((u8 *)frame + size);

	csa->id = IEEE80211_IE_CSA;
	csa->len = 3;

	/* New Channel */
	csa->channel = priv->set_csa_channel ? priv->csa_channel : 40;

	/* Mode:
	 * 1 - No more Tx allowed until on the new channel
	 * 0 - No Tx rerequiement on the receiving client
	 */
	csa->mode = 1;

	/* Channel Switch Count:
	 * >1 - Number of beacon transmission intervals until switch.
	 * 1  - Occur immediately before next TBTT.
	 * 0  - Occur immediately
	 */
	csa->count = 0;

	return size + sizeof(*csa);
}

int init_deauth(struct priv *priv, struct ieee80211_deauth *frame)
{
	if (!priv->set_bss || !priv->set_dst || !priv->set_src) {
		fprintf(stderr, "--bss, --dst, and --src "
			"must all be set.\n");
		return 0;
	}
	init_ieee80211_hdr_3addr(priv, &frame->header,
				 IEEE80211_FTYPE_MGMT | 
				 IEEE80211_STYPE_DEAUTH);
	frame->reason = 5;

	return sizeof(*frame);
}

int set_iface_flags(const char *ifname, int flags)
{
	struct ifreq ifr;
	int sock, rc;

	sock = socket(PF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		rc = errno;
		fprintf(stderr, "Error: socket() failed reason (%d): %s\n",
			errno, strerror(errno));
		errno = rc;
		return -rc;
	}

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
	ifr.ifr_flags = flags & 0xffff;
	rc = ioctl(sock, SIOCSIFFLAGS, &ifr);
	if (rc < 0) {
		rc = errno;
		fprintf(stderr, "Error: ioctl() failed reason (%d): %s\n",
			errno, strerror(errno));
		errno = rc;
		close(sock);
		return -rc;
	}

	close(sock);

	return 0;
}

int get_iface_flags(const char *ifname, short *flags) 
{
	struct ifreq ifr;
	int sock, rc;

	*flags = 0;

	sock = socket(PF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		rc = errno;
		fprintf(stderr, "Error: socket() failed reason (%d): %s\n",
			errno, strerror(errno));
		errno = rc;
		return -rc;
	}

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
	rc = ioctl(sock, SIOCGIFFLAGS, &ifr);
	if (rc < 0) {
		rc = errno;
		fprintf(stderr, "Error: ioctl() failed reason (%d): %s\n",
			errno, strerror(errno));
		errno = rc;
		close(sock);
		return -rc;
	}

	close(sock);

	*flags = ifr.ifr_flags;
	
	return 0;
}

int open_iface(struct priv *priv)
{
	struct sockaddr_ll sll_addr = {
		.sll_family = AF_PACKET,
		.sll_ifindex = -1,
		.sll_protocol = htons(ETH_P_80211_RAW),
		.sll_pkttype = PACKET_OTHERHOST,
	};
	struct ifreq ifr = {
		.ifr_name = "",
	};
	int err;
	socklen_t err_len = sizeof(err);
	int rc;
	short flags;

	if (priv->set_rtap)
		sll_addr.sll_hatype = ARPHRD_IEEE80211_RADIOTAP;
	else
		sll_addr.sll_hatype = ARPHRD_IEEE80211;

	strncpy(ifr.ifr_name, priv->ifname, sizeof(ifr.ifr_name));

	rc = get_iface_flags(priv->ifname, &flags);
	if (rc < 0) {
		fprintf(stderr, 
			"Error: get_iface_flags failed reason (%d): %s\n",
			errno, strerror(errno));
		return 1;
	}

	if (!(flags & IFF_UP)) {
		if (priv->set_no_configure) {
			fprintf(stderr, "Error: %s not UP and "
				"--no-configure was indicated.\n",
				priv->ifname);
			return -1;
		}
		if (!priv->set_quiet)
			fprintf(stdout, "Bringing %s IFF_UP.\n", 
				priv->ifname);
		rc = set_iface_flags(priv->ifname, flags | IFF_UP);
		if (rc < 0) {
			fprintf(stderr, 
				"Error: set_iface_flags on %s failed reason "
				"(%d): %s\n",
				priv->ifname, errno, strerror(errno));
			return 1;
		}

		priv->app_configured = 1;
	}

	priv->sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_80211_RAW));
	if (priv->sock == -1) {
		fprintf(stderr, "Error: socket() failed reason (%d): %s\n",
			errno, strerror(errno));
		return 1;
	}

	if (ioctl(priv->sock, SIOCGIFINDEX, &ifr) == -1) {
		fprintf(stderr, "Error: ioctl() failed reason (%d): "
			"%s\n", errno, strerror(errno));
		return 1;
	}

	sll_addr.sll_ifindex = ifr.ifr_ifindex;

	if (!priv->set_quiet)
		fprintf(stdout, "using: %s\n", ifr.ifr_name);

	/* Bind to the appropiate interface */
	if (bind(priv->sock, (struct sockaddr *)&sll_addr, 
		 sizeof(sll_addr))) {
		fprintf(stderr, "Error: bind() failed reason (%d): %s\n",
			errno, strerror(errno));
		return 1;
	}

	/* Verify no errors on socket */
	if (getsockopt(priv->sock, SOL_SOCKET, SO_ERROR, &err, &err_len)) {
		fprintf(stderr, "Error: getsockopt() reason (%d): %s\n",
			errno, strerror(errno));
		return 1;
	}

	return 0;
}

int close_iface(struct priv *priv)
{
	short flags;
	int rc;

	close(priv->sock);
	
	if (priv->app_configured) {
		rc = get_iface_flags(priv->ifname, &flags);
		if (!rc && (flags & IFF_UP)) {
			if (!priv->set_quiet)
				fprintf(stdout, "Removing %s IFF_UP.\n", 
					priv->ifname);
			rc = set_iface_flags(priv->ifname, flags & ~IFF_UP);
			if (rc < 0) {
				fprintf(stderr, 
					"Error: set_iface_flags on %s failed "
					"reason (%d): %s\n",
					priv->ifname, errno, strerror(errno));
			}
		}
	}

	return 0;
}

#ifndef offsetof
#define offsetof(x, y) ((size_t)&((x *)0)->y)
#endif

void send_frame(struct priv *priv, struct sim_frame *frame, int len)
{
	int sent;

	if (priv->set_rtap) {
		len += offsetof(struct sim_frame, type);
		sent = send(priv->sock, frame, len, MSG_NOSIGNAL);
	} else {
		sent = send(priv->sock, &frame->type, len, MSG_NOSIGNAL);
	}
	if (sent == -1) {
		fprintf(stderr, "send failed for %d bytes: %s (%d)\n",
			len, strerror(errno),
			errno);
		return;
	}

	if (sent < len) {
		fprintf(stderr, "Short packet transmitted.\n");
		return;
	}
}

enum {
	CSA = 0,
	DEAUTH = 1,
	BEACON = 2,
	DATA = 3,
};

int main(int argc, char *argv[])
{
	struct priv priv;
	int rc = 0, pos, i = 0;
	struct sim_frame frame;
	int len;
	int frame_type = DEAUTH;
	struct timeval tv;

	gettimeofday(&tv, NULL);
	srand(tv.tv_usec);

	memset(&priv, 0, sizeof(priv));
	while (i < argc - 1) {
		i++;

		if (argv[i][0] != '-' || argv[i][1] != '-')
			return usage(argv, i);

		pos = param_check(&argv[i][2], "no-configure", PARAM_NONE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.set_no_configure = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "i", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.ifname = strdup(&argv[i][2 + pos]);
			priv.set_ifname = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "quiet", PARAM_NONE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.set_quiet = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "count", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.set_count = 1;
			priv.count = strtol(&argv[i][2 + pos], NULL, 0);
			continue;
		}

		pos = param_check(&argv[i][2], "size", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.set_size = 1;
			priv.size = strtol(&argv[i][2 + pos], NULL, 0);
			if (priv.size < IEEE80211_3ADDR_LEN || priv.size > 2346) 
				return usage(argv, i);
			priv.size -= IEEE80211_3ADDR_LEN;
			continue;
		}

		pos = param_check(&argv[i][2], "rtap", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.set_rtap = (strtol(&argv[i][2 + pos], NULL, 0) 
					 == 1) ? 1 : 0;
			continue;
		}

		pos = param_check(&argv[i][2], "type", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			frame_type = strtol(&argv[i][2 + pos], NULL, 0);
			priv.set_type = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "csa-channel", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.csa_channel = strtol(&argv[i][2 + pos], NULL, 0);
			priv.set_csa_channel = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "channel", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.channel = strtol(&argv[i][2 + pos], NULL, 0);
			priv.set_channel = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "src", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			if (parse_mac(&argv[i][2 + pos], priv.src))
				return usage(argv, i);
			priv.set_src = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "dst", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			if (parse_mac(&argv[i][2 + pos], priv.dst))
				return usage(argv, i);
			priv.set_dst = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "bss", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			if (parse_mac(&argv[i][2 + pos], priv.bss))
				return usage(argv, i);
			priv.set_bss = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "help", PARAM_NONE);
		if (pos < 0)
			return usage(argv, i);
		if (pos)
			return usage(argv, 0);

		pos = param_check(&argv[i][2], "version", PARAM_NONE);
		if (pos < 0)
			return usage(argv, i);
		if (pos)
			return version();

		return usage(argv, i);
	}

	if (!priv.set_count)
		priv.count = 1;

	if (!priv.set_ifname) {
		if (priv.set_rtap)
			priv.ifname = strdup("rtap0");
		else
			priv.ifname = strdup("wifi0");
	}

	if (priv.set_rtap)
		if (!init_radiotap(&priv, &frame.rt))
			return -1;

	if (!priv.set_type) {
		fprintf(stderr, "--type must be set.\n");
		return 0;
	}


	rc = open_iface(&priv);
	if (rc) 
		return -rc;

	for (i = 0; i < priv.count; i++) {
		switch (frame_type) {
		case CSA:
			len = init_csa(&priv, &frame.type.beacon);
			break;
		case DEAUTH:
			len = init_deauth(&priv, &frame.type.deauth);
			break;
		case BEACON:
			len = init_beacon(&priv, &frame.type.beacon);
			break;
		case DATA:
			len = init_data(&priv, &frame.type.data);
			break;
		}

		if (len == 0)
			return -1;

		send_frame(&priv, &frame, len);
	}
	rc = close_iface(&priv);

	return -rc;
}
